#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>

#ifdef __DJGPP__
 #include <dir.h>
#else
 #include "ffirst.h"
 #include <stdlib.h>
 #include <time.h>
#endif

#ifndef __LINUX__
 #include <io.h>
#endif

#if defined (__OS2__) || defined (__DOS__)
 #include <dos.h>
#endif

#include <sys/stat.h>
#include "msxboot2.h"

#define VERSION "3.1"

typedef unsigned char byte;
typedef unsigned short word;

typedef struct {
  char name[9];
  char ext[4];
  int size;
  int hour,min,sec;
  int day,month,year;
  int first;
  int pos;
  int attr;
} fileinfo;

byte *dskimage=NULL;
byte *fat;
byte *direc;
byte *cluster;
int sectorspercluster,sectorsperfat,numberoffats,reservedsectors;
int bytespersector,direlements,fatelements,totalsectors;
int availsectors;
int bytespercluster;

char **__crt0_glob_function (char *_argument) {
  return NULL;
}

/**************************/
/* Get File Date          */
/* Return file size in KB */
/**************************/
time_t getFileDate(char * filename) {
#if defined(__WATCOMC__) || defined(__GNUC__) || defined(__LINUX__)
   int handle;
   int rc;
   struct stat data;
   handle = open( filename, O_RDONLY | O_BINARY);
   if( handle == -1 ) return 0;
   rc = fstat(handle,&data);
   close( handle );
   if (rc==0) return( data.st_mtime );
   else return time(NULL);
#else
    return time(NULL);
#endif
   }


/**************************/
/* Get File Size          */
/* Return file size in KB */
/**************************/
int getFileSize(char * filename) {
#if defined(__WATCOMC__) || defined(__GNUC__) || defined(__LINUX__)
   int handle;
   int size_of_file;
   handle = open( filename, O_RDONLY | O_BINARY);
   if( handle == -1 ) return 0;
   size_of_file = (int)(filelength(handle)/1024);
   close( handle );
   return( size_of_file );
#else
   FILE *fp;
   int size_of_file;
   fp = fopen( filename, "r" );
   if (fp == NULL) return 0;
   fseek( fp, 0L, SEEK_END );
   size_of_file = (int)( ftell( fp ) / 1024 );
   fclose( fp );
   return( size_of_file );
#endif
   }


/***********************************************/
/* Adjust Boot Sector Accordingly to Disk Size */
/***********************************************/
void setDiskBoot(byte * dskimage, int dsize) {
   if (dsize == 180) {
       *(dskimage+0x0D) = 1; // 1 sector per cluster
       *(dskimage+0x11) = 64; // only 64 directory entries
       *(dskimage+0x13) = 0x68; // 360 sectors on disk
       *(dskimage+0x14) = 0x01;
       *(dskimage+0x15) = 0xFC; // Disk Type
       *(dskimage+0x16) = 2; // Only two sectors for FAT
       *(dskimage+0x1A) = 1; // Single side disk
       }
    else if (dsize == 360) {
       *(dskimage+0x0D) = 2; // 2 sector per cluster
       *(dskimage+0x11) = 112; // Full 112 directory entries
       *(dskimage+0x13) = 0xD0; // 720 sectors on disk
       *(dskimage+0x14) = 0x02;
       *(dskimage+0x15) = 0xFD; // Disk Type
       *(dskimage+0x16) = 2; // Only two sectors for FAT
       *(dskimage+0x1A) = 2; // Double side disk
       }
    else if (dsize == 8160) { // 8160 disk
       *(dskimage+0x0D) = 4; // 4 sector per cluster
       *(dskimage+0x11) = 224; // Full 224 directory entries 
       *(dskimage+0x13) = 0xC0; // 16320 sectors on disk
       *(dskimage+0x14) = 0x3F;
       *(dskimage+0x15) = 0xF8; // Disk Type
       *(dskimage+0x16) = 12; // 12 sectors for FAT
       *(dskimage+0x1A) = 2; // Double side disk
       }
    else if (dsize == 16320) { // 16320 disk
       *(dskimage+0x0D) = 8; // 8 sectors per cluster
       *(dskimage+0x11) = 224; // Full 224 directory entries 
       *(dskimage+0x13) = 0x80; // 32640 sectors on disk  (4080 clusters)
       *(dskimage+0x14) = 0x7F;
       *(dskimage+0x15) = 0xF8; // Disk Type
       *(dskimage+0x16) = 12; // 12 sectors for FAT
       *(dskimage+0x1A) = 2; // Double side disk
       }
    else if (dsize == 32644) { // 32644 disk
       *(dskimage+0x0D) = 16; // 16 sector per cluster
       *(dskimage+0x11) = 224; // Full 224 directory entries 
       *(dskimage+0x13) = 0x08; // 65288 sectors on disk (4080 clusters)
       *(dskimage+0x14) = 0xFF;
       *(dskimage+0x15) = 0xF8; // Disk Type
       *(dskimage+0x16) = 12; // 12 sectors per fat (3 sectors for each 1024 clusters)
       *(dskimage+0x1A) = 2; // Double side disk
       }
    else { // 720 disk
       *(dskimage+0x0D) = 2; // 2 sector per cluster
       *(dskimage+0x11) = 112; // Full 112 directory entries
       *(dskimage+0x13) = 0xA0; // 1440 sectors on disk
       *(dskimage+0x14) = 0x05;
       *(dskimage+0x15) = 0xF9; // Disk Type
       *(dskimage+0x16) = 3; // Three sectors for FAT
       *(dskimage+0x1A) = 2; // Double side disk
       }
   /* Set DISK Image Name */
   strcpy((dskimage+0x03),"DSKTool");
   *(dskimage+0x0A) = '1';
   }



/***********************************************/
/* Load a disk image... or creates one, if file*/
/* does not exist and a size was provided      */
/* Output: size of disk created. 0 if error.   */
/***********************************************/
int load_dsk (char *name, int error, int dsize) {
   int file;
   int dstmp;

   //printf("Requested Disk Size: %d\n",dsize);
   dstmp = getFileSize(name);
   //printf("Real Disk Size:     %d\n",dstmp);
   if (dstmp != 0) dsize = dstmp;
   if (dsize == 0) return 0;

   printf("Disk Size: %dKB\n",dsize);

   dskimage=(byte *) malloc (dsize*1024);
   if (dskimage==NULL) {
      printf ("not enought memory\n");
      exit (2);
      }
   file=open (name,O_BINARY|O_RDONLY);
   if (file<0) {
      if (error) {
         printf ("Error in .DSK file\n");
         exit (2);
         }
      memset (dskimage,0,dsize*1024);
      memcpy (dskimage,msxboot,512);
      setDiskBoot(dskimage,dsize);
      }
   else {
      read (file,dskimage,dsize*1024);
      close (file);
      }
   bytespersector=*(word *)(dskimage+0x0B);
   sectorspercluster=*(dskimage+0x0D);
   reservedsectors=*(word *)(dskimage+0x0E);
   numberoffats=*(dskimage+0x10);
   totalsectors=*(word *)(dskimage+0x13);
   sectorsperfat=*(word *)(dskimage+0x16);
   direlements=*(word *)(dskimage+0x11);
   fat=dskimage+bytespersector*reservedsectors;
   direc=fat+bytespersector*(sectorsperfat*numberoffats);
   cluster=direc+direlements*32;
   availsectors=totalsectors-reservedsectors-(sectorsperfat*numberoffats);
   availsectors-=((direlements*32)/bytespersector);
   fatelements=availsectors/sectorspercluster;
   bytespercluster = bytespersector * sectorspercluster;  /* On 720K, block is 1024 */

   if (file<0) {
      fat[0]=0xF9;
      fat[1]=0xFF;
      fat[2]=0xFF;
      }
   return dsize;
   }


int next_link (int link) {
  int pos;

  pos=(link>>1)*3;
  if (link&1)
    return (((int)(fat[pos+2]))<<4)+(fat[pos+1]>>4);
  else
    return (((int)(fat[pos+1]&0xF))<<8)+fat[pos];
}

int remove_link (int link) {
  int pos;
  int current;

  pos=(link>>1)*3;
  if (link&1) {
    current=(((int)(fat[pos+2]))<<4)+(fat[pos+1]>>4);
    fat[pos+2]=0;
    fat[pos+1]&=0xF;
    return current;
  }
  else  {
    current=(((int)(fat[pos+1]&0xF))<<8)+fat[pos];
    fat[pos]=0;
    fat[pos+1]&=0xF0;
    return current;
  }
}

void store_fat (int link, int next) {
  int pos;

  pos=(link>>1)*3;
  if (link&1) {
    fat[pos+2]=next>>4;
    fat[pos+1]&=0xF;
    fat[pos+1]|=(next&0xF)<<4;
  }
  else  {
    fat[pos]=next&0xFF;
    fat[pos+1]&=0xF0;
    fat[pos+1]|=next>>8;
  }
}

fileinfo *getfileinfo (int pos)
{
  fileinfo *file;
  byte *dir;
  int i;

  dir=direc+pos*32;
  if (*dir<0x20 || *dir>=0x80) return NULL;

  file=(fileinfo *) malloc (sizeof (fileinfo));
  if (file==NULL)
  {
    return NULL;
  }

  for (i=0; i<8; i++)
    file->name[i]=dir[i]==0x20?0:dir[i];
  file->name[8]=0;

  for (i=0; i<3; i++)
    file->ext[i]=dir[i+8]==0x20?0:dir[i+8];
  file->ext[3]=0;

  file->size=*(int *)(dir+0x1C);

  i=*(word *)(dir+0x16);
  file->sec=(i&0x1F)<<1;
  file->min=(i>>5)&0x3F;
  file->hour=i>>11;

  i=*(word *)(dir+0x18);
  file->day=i&0x1F;
  file->month=(i>>5)&0xF;
  file->year=1980+(i>>9);

  file->first=*(word *)(dir+0x1A);
  file->pos=pos;
  file->attr=*(dir+0xB);

  return file;
}

int bytes_free (void)
{
  int i,avail=0;

  for (i=2; i<2+fatelements; i++)
  {
    if (!next_link (i)) avail++;
  }
  return avail * bytespercluster;
}

void list_dsk (void)
{
  int i;
  fileinfo *file;
  char name[20];
  char date[30];
  char ftime[30];
  char size[30];
  int fcount;
  long tsize;

  fcount=0;
  tsize=0;
  for (i=0; i<8; i++) name[i]=dskimage[3+i];

  name[8]=0;
  printf ("Name of volume: %s\n\n",name);
  for (i=0; i<direlements; i++) {
    file=getfileinfo (i);
    if (file!=NULL) {
      fcount++;
      if (file->ext[0])
        sprintf (name,"%s.%s",file->name,file->ext);
      else
        strcpy (name,file->name);
      sprintf (size,"%7d",file->size);
      tsize+=file->size;
      if (file->attr&0x8) strcpy (size,"  <VOL>");
      if (file->attr&0x10) strcpy (size,"  <DIR>");
      sprintf (date,"%d/%02d/%d",file->day,file->month,file->year);
      sprintf (ftime,"%d:%02d:%02d",file->hour,file->min,file->sec);
      printf ("%-13s %s %10s %8s\n",name,size,date,ftime);
      free (file);
    }
  }
  printf ("\n%10ld bytes used by %5d file(s)\n",tsize,fcount);
  printf ("%10d bytes free in %5d cluster(s)\n",bytes_free(),bytes_free()/bytespercluster);
}

/* Compare a file on disk image with a file outside it (name) */
/* Output: 1 if they match. 0 Otherwise. */
int match (fileinfo *file, char *name) {
   char *p=file->name;
   int status=0,i;

   //printf ("Comparing: %s.%s to %s\n",file,file->ext,name);
   /* Verify if file on disk has the same */
   /* name of the file do be added/removed */
   for (i=0; i<8; i++) {
      if (!*name) break; /* Empty name, test outside */
      if (*name=='*') {  /* Name is *, so they *are* equal */
         status=1;
         name++;
         break;
         }
      if (*name=='.') break; /* Reached extension? Test outside */
      /* Normal name... compare... return 0 if they are different */
      if (toupper (*name++)!=toupper (*p++)) return 0;
      }
   /* If the names are not the same size, they are not equal! */
   if (!status && i<8 && *p!=0) return 0;
   /* Jump to point... */
   /* Equal so far... */
   /* Now tests the extension */
   p=file->ext;
   /* If both have no extensions, they are equal... */
   if (!*name && !*p) return 1;
   /* Long filename being added */
   if (i==8) while (*name && *name!='.') name++;
   /* Test extension now... */
   if (*name++!='.') return 0; /* if name ended and file name not... not equal */
   /* Both have extensions! */
   for (i=0; i<3; i++) {
      if (!*name) break; /* Empty extension, test outside */
      if (*name=='*') return 1; /* always equal when extension is * */
      /* If one letter is different, they are not equal! */
      if (toupper (*name++)!=toupper (*p++)) return 0;
      }
   /* If extensions are not the same size, filenames are not equal! */
   if (i<3 && *p!=0) return 0;
   /* If everthing goes ok until now, the files are equal */
   return 1;
   }

void parse_tree (char *name, void (*action)(fileinfo *)) {
   int i;
   fileinfo *file;

   for (i=0; i<direlements; i++) {
      if ((file=getfileinfo (i))!=NULL) {
         if (match (file,name)) {
            action (file);
            }
         free (file);
         }
      }
   }

void parse_dsk (int argc, char **argv, void (*action)(fileinfo *)) {
  int i;

  if (argc==3)
    parse_tree ("*.*",action);
  else
    for (i=3; i<argc; i++)
      parse_tree (argv[i],action);
}

void extract (fileinfo *file) {
  byte *buffer,*p;
  int fileid;
  char name[20];
  int current;

  printf ("extracting %s.%s\n",file->name,file->ext);
  buffer=(byte *) malloc ((file->size+(bytespercluster-1))&(~(bytespercluster-1)));
  if (buffer!=NULL)
  {
      memset (buffer,0x1a,file->size);
      sprintf (name,"%s.%s",file->name,file->ext);
      fileid=open (name,O_BINARY|O_WRONLY|O_CREAT,S_IRUSR|S_IWUSR);
      current=file->first;
      p=buffer;
      do {
        memcpy (p,cluster+(current-2)*bytespercluster,bytespercluster);
        p+=bytespercluster;
        current=next_link (current);
      } while (current!=0xFFF);
      write (fileid,buffer,file->size);
      close (fileid);
      free (buffer);
  }
}

void wipe (fileinfo *file) {
  int current;

  current=file->first;
  do {
    current=remove_link (current);
  } while (current!=0xFFF);
  direc[file->pos*32]=0xE5;
}

void deleteFile (fileinfo *file) {
  printf ("deleting %s.%s\n",file->name,file->ext);
  wipe (file);
}

void flush_dsk (char *name, int dsize)
{
  int file;

  memcpy (fat+bytespersector*sectorsperfat,fat,bytespersector*sectorsperfat);
  file=open (name,O_BINARY|O_WRONLY|O_CREAT,S_IWUSR|S_IRUSR);
  if (file>=0)
  {
    write (file,dskimage,dsize*1024);
    close (file);
  }
}

int get_free (void) {
  int i;

  for (i=2; i<2+fatelements; i++)
     if (!next_link (i)) return i;
  return -1;
}

int get_next_free (void) {
  int i,status=0;

  for (i=2; i<2+fatelements; i++)
    if (!next_link (i))
    {
      if (status) { return i; } else { status=1; }
    }
  return -1;
}

void add_single_file (char *name, char *pathname) {
  int i,total;
  int found=0;
  fileinfo *file=NULL;
  int fileid;
  byte *buffer;
  int size;
  int first;
  int current;
  int next;
  int pos;
  char *p;
  int filesizespace;
  char fullname[250];
  byte * tmp_buffer;
#ifdef __DJGPP__
  struct time ti;
  struct date da;
#else
  struct tm * ti;
  time_t tmday;
#endif

  strcpy (fullname,pathname);
  strcat (fullname,name);
  fileid=open (fullname,O_BINARY|O_RDONLY);
  if (fileid>=0)
  {
          for (i=0; i<direlements; i++)
          {
            if ((file=getfileinfo (i))!=NULL)
            {
              if (match (file,name))
              {
                found=1;
                wipe (file);
              }
              free (file);
            }
          }

          if ((size=filelength(fileid))>bytes_free())
          {
            printf ("disk full\n");
            exit (4);
          }

          if (found)
            printf ("updating %s\n",name);
          else
            printf ("  adding %s\n",name);

          for (i=0; i<direlements; i++)
          {
            if (direc[i*32]<0x20 || direc[i*32]>=0x80) break;
          }

          if (i==direlements)
          {
            printf ("directory full\n");
            exit (6);
          }
          pos=i;

          filesizespace = (size+((size>0)?0:1)+(bytespercluster-1)) & (~(bytespercluster-1));
          //printf ("*** Arquivo de %d bytes, alocando... %d bytes\n", size, filesizespace);
          buffer=(byte *) malloc ( filesizespace );
          if (buffer==NULL)
          {
            printf ("Not enough memory!\n");
            exit (2);
          }

          read (fileid,buffer,size);
          close (fileid);

          total=(size+bytespercluster)/bytespercluster;
          current=first=get_free();
          if (current < 0) {
             printf ("Internal error\n");
             exit (5);
             }

          tmp_buffer=buffer;
          for (i=0; i<total;)
          {
            memcpy (cluster+(current-2)*bytespercluster,tmp_buffer,bytespercluster);
            tmp_buffer+=bytespercluster;
            if (++i==total)
              next=0xFFF;
            else {
               next=get_next_free ();
               if (next < 0) {
                  printf ("Internal error\n");
                  exit (5);
                  }
               }
            store_fat (current,next);
            current=next;
          }

          memset (direc+pos*32,0,32);
          memset (direc+pos*32,0x20,11);
          i=0;
          for (p=name;*p;p++)
          {
            if (*p=='.')
            {
              i=8;
              continue;
            }
            direc[pos*32+i++]=toupper (*p);
            /* If filename is bigger than 8 chars... continue */
            if (i==8 && *(p+1)!='.') while (*(p+1)!='.' && *(p+1)!='\0') p++;
            /* Extension is bigger than 3 chars... break. */
            if (i>10) break;
          }

          *(word *)(direc+pos*32+0x1A)=first;
          *(int *)(direc+pos*32+0x1C)=size;
#ifdef __DJGPP__
          gettime (&ti);
          getdate (&da);
          *(word *)(direc+pos*32+0x16)=
            (ti.ti_sec>>1)+(ti.ti_min<<5)+(ti.ti_hour<<11);
          *(word *)(direc+pos*32+0x18)=
            (da.da_day)+(da.da_mon<<5)+((da.da_year-1980)<<9);
#else
          tmday = (time_t)getFileDate(fullname);
          ti = localtime(&tmday);
          *(word *)(direc+pos*32+0x16)=
            (ti->tm_sec>>1)+(ti->tm_min<<5)+(ti->tm_hour<<11);
          *(word *)(direc+pos*32+0x18)=
            (ti->tm_mday)+((ti->tm_mon+1)<<5)+((ti->tm_year-80)<<9);
#endif

          free (buffer);
  }
}

void add_files (char *name) {
   int status;
   char *temp1,*temp2;
   char dir[200];
#ifdef __DJGPP__
   struct ffblk finfo;
#else
   struct dirdata finfo;
#endif

   status=findfirst (name,&finfo,FA_ARCH|FA_RDONLY);

   temp1=NULL;
   temp2=name;
   while ((temp2=strstr (temp2,"\\"))!=NULL) {
      temp1=temp2;
      temp2++;
      }
   if (temp1!=NULL) {
      memset (dir,0,200);
      memcpy (dir,name,temp1-name);
      strcat (dir,"\\");
      }
   else {
      *dir=0;
      }

   while (!status) {
#ifdef __DJGPP__
      add_single_file(finfo.ff_name,dir);
#else
      add_single_file(finfo.finfo->d_name,dir);
#endif
      status=findnext(&finfo);
      //printf ("Status: %d\n",status);
      }
   }

void add_to_dsk (int argc, char **argv) {
  int i;

  for (i=3; i<argc; i++)
    add_files (argv[i]);
}

int main (int argc, char **argv) {
  int dsize;

  printf ("DSK Tool v%s\n", VERSION);
  printf ("(B)1998, Ricardo Bittencourt\n(B)2006, SLotman / Icon Games\n");
  printf ("BUGright 2010-2018 by DJC / Amusement Factory\n");
  printf ("This file is under GNU GPL, read COPYING for details\n\n");

  if (argc<3) {
    printf ("Usage: DSKTOOL <cmd>[:disk_size] <archive> [files]\n\n");
    printf ("cmds:");
    printf ("\tD\tdeleteFile files from .DSK\n");
    printf ("\tE\textract files from .DSK\n");
    printf ("\tL\tlist contents of .DSK\n");
    printf ("\tA[size]\tadd files to .DSK\n");
    printf ("\nsizes:");
    printf ("\t180\t180 KB Disk\n");
    printf ("\t360\t360 KB Disk\n");
    printf ("\t720\t720 KB Disk (default)\n");
    printf ("\t8\t8 MB Hard Disk\n");
    printf ("\t16\t16 MB Hard Disk\n");
    printf ("\t32\t32 MB Hard Disk\n");
    printf ("\nexamples:\n");
    printf ("\tDSKTOOL L TALKING.DSK\n");
    printf ("\tDSKTOOL E TALKING.DSK FUZZ*.*\n");
    printf ("\tDSKTOOL D TALKING.DSK *.BAS *.BIN\n");
    printf ("\tDSKTOOL A TALKING.DSK MSXDOS.SYS COMMAND.COM\n");
    printf ("\tDSKTOOL A180 TALKING.DSK MSXDOS.SYS COMMAND.COM\n");
    exit (1);
  }

  dsize = 720;
  if(argv[1][1]=='1' && argv[1][2]=='8') dsize = 180;
  else if(argv[1][1]=='3' && argv[1][2]=='6') dsize = 360;
  else if(argv[1][1]=='8') dsize = 8160;
  else if(argv[1][1]=='1' && argv[1][2]=='6') dsize = 16320;
  else if(argv[1][1]=='3' && argv[1][2]=='2') dsize = 32644;

  switch (toupper (argv[1][0])) {
    case 'L':
      dsize = load_dsk (argv[2],1,0);
      if (dsize==0) break;
      list_dsk();
      break;
    case 'E':
      dsize = load_dsk (argv[2],1,0);
      if (dsize==0) break;
      parse_dsk (argc, argv, extract);
      break;
    case 'D':
      dsize = load_dsk (argv[2],1,0);
      if (dsize==0) break;
      parse_dsk (argc, argv, deleteFile);
      flush_dsk (argv[2], dsize);
      break;
    case 'A':
      dsize = load_dsk (argv[2], 0, dsize);
      if (dsize==0) break;
      add_to_dsk (argc, argv);
      flush_dsk (argv[2], dsize);
      break;
    default:
      printf ("Command not supported\n");
      exit (3);
  }
  if (dsize==0) {
     printf ("File %s does not exist.\nNothing to do.\n",argv[2]);
     exit (6);
     }
  return 0;
}
